using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Domain.Entities;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using Microsoft.Extensions.Localization;

namespace cuqmbr.TravelGuide.Application.Employees.Commands.AddEmployee;

public class AddEmployeeCommandHandler :
    IRequestHandler<AddEmployeeCommand, EmployeeDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;
    private readonly IStringLocalizer _localizer;

    public AddEmployeeCommandHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        IStringLocalizer localizer)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _localizer = localizer;
    }

    public async Task<EmployeeDto> Handle(
        AddEmployeeCommand request,
        CancellationToken cancellationToken)
    {
        var parentEntity = await _unitOfWork.CompanyRepository.GetOneAsync(
            e => e.Guid == request.CompanyGuid, cancellationToken);

        if (parentEntity == null)
        {
            throw new NotFoundException(
                $"Parent entity with Guid: {request.CompanyGuid} not found.");
        }

        var entity = await _unitOfWork.EmployeeRepository.GetOneAsync(
            e =>
                e.FirstName == request.FirstName &&
                e.LastName == request.LastName &&
                e.Patronymic == request.Patronymic &&
                e.Sex == request.Sex &&
                e.BirthDate == request.BirthDate &&
                e.CompanyId == parentEntity.Id,
            cancellationToken);

        if (entity != null)
        {
            throw new DuplicateEntityException();
        }

        entity = new Employee()
        {
            FirstName = request.FirstName,
            LastName = request.LastName,
            Patronymic = request.Patronymic,
            Sex = request.Sex,
            BirthDate = request.BirthDate,
            Documents = request.Documents.Select(
                d => new EmployeeDocument()
                {
                    DocumentType = d.DocumentType,
                    Information = d.Information
                })
                .ToArray(),
            Company = parentEntity
        };

        entity = await _unitOfWork.EmployeeRepository.AddOneAsync(
            entity, cancellationToken);

        await _unitOfWork.SaveAsync(cancellationToken);
        _unitOfWork.Dispose();

        return _mapper.Map<EmployeeDto>(entity);
    }
}
