using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using Microsoft.Extensions.Localization;
using cuqmbr.TravelGuide.Domain.Entities;

namespace cuqmbr.TravelGuide.Application.Employees.Commands.UpdateEmployee;

public class UpdateEmployeeCommandHandler :
    IRequestHandler<UpdateEmployeeCommand, EmployeeDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;
    public IStringLocalizer _localizer { get; set; }

    public UpdateEmployeeCommandHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        IStringLocalizer localizer)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _localizer = localizer;
    }

    public async Task<EmployeeDto> Handle(
        UpdateEmployeeCommand request,
        CancellationToken cancellationToken)
    {
        var parentEntity = await _unitOfWork.CompanyRepository.GetOneAsync(
            e => e.Guid == request.CompanyGuid, cancellationToken);

        if (parentEntity == null)
        {
            throw new NotFoundException(
                $"Parent entity with Guid: {request.CompanyGuid} not found.");
        }


        var entity = await _unitOfWork.EmployeeRepository.GetOneAsync(
            e =>
                e.FirstName == request.FirstName &&
                e.LastName == request.LastName &&
                e.Patronymic == request.Patronymic &&
                e.Sex == request.Sex &&
                e.BirthDate == request.BirthDate &&
                e.CompanyId == parentEntity.Id &&
                e.Guid != request.Guid,
            cancellationToken);

        if (entity != null)
        {
            throw new DuplicateEntityException();
        }


        entity = await _unitOfWork.EmployeeRepository.GetOneAsync(
            e => e.Guid == request.Guid, e => e.Documents, cancellationToken);

        if (entity == null)
        {
            throw new NotFoundException();
        }


        entity.Guid = request.Guid;
        entity.FirstName = request.FirstName;
        entity.LastName = request.LastName;
        entity.Patronymic = request.Patronymic;
        entity.Sex = request.Sex;
        entity.BirthDate = request.BirthDate;
        entity.CompanyId = parentEntity.Id;

        entity.Company = parentEntity;


        var requestEmployeeDocuments = request.Documents.Select(
            d => new EmployeeDocument()
            {
                DocumentType = d.DocumentType,
                Information = d.Information
            });

        var commonEmployeeDocuments = entity.Documents.IntersectBy(
            requestEmployeeDocuments.Select(
                ed => (ed.DocumentType, ed.Information)),
            ed => (ed.DocumentType, ed.Information));

        var newEmployeeDocuments = requestEmployeeDocuments.ExceptBy(
            entity.Documents.Select(ed => (ed.DocumentType, ed.Information)),
            ed => (ed.DocumentType, ed.Information));

        var combinedEmployeeDocuments = commonEmployeeDocuments.UnionBy(
            newEmployeeDocuments, ed => (ed.DocumentType, ed.Information));

        entity.Documents = combinedEmployeeDocuments.ToList();


        entity = await _unitOfWork.EmployeeRepository.UpdateOneAsync(
            entity, cancellationToken);

        await _unitOfWork.SaveAsync(cancellationToken);
        _unitOfWork.Dispose();

        return _mapper.Map<EmployeeDto>(entity);
    }
}
