using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Domain.Entities;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Enums;
using FluentValidation.Results;
using Microsoft.Extensions.Localization;

namespace cuqmbr.TravelGuide.Application.Payments.LiqPay
    .TicketGroups.Commands.GetPaymentLink;

public class GetPaymentLinkCommandHandler :
    IRequestHandler<GetPaymentLinkCommand, PaymentLinkDto>
{
    private readonly UnitOfWork _unitOfWork;

    private readonly CurrencyConverterService _currencyConverterService;

    private readonly LiqPayPaymentService _liqPayPaymentService;

    private readonly IStringLocalizer _localizer;

    public GetPaymentLinkCommandHandler(
        UnitOfWork unitOfWork,
        CurrencyConverterService currencyConverterService,
        LiqPayPaymentService liqPayPaymentService,
        IStringLocalizer localizer)
    {
        _unitOfWork = unitOfWork;
        _currencyConverterService = currencyConverterService;
        _liqPayPaymentService = liqPayPaymentService;
        _localizer = localizer;
    }

    public async Task<PaymentLinkDto> Handle(
        GetPaymentLinkCommand request,
        CancellationToken cancellationToken)
    {
        // Check whether provided vehicle enrollments are present in datastore.
        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            if (vehicleEnrollmentGuids.Count() > vehicleEnrollments.Count)
            {
                throw new NotFoundException();
            }
        }


        // Check whether provided arrival and departure address guids
        // are used in provided vehicle enrollment and
        // and are in the correct order.
        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            var routeAddressGuids =
                request.Tickets.Select(t => t.DepartureRouteAddressGuid).Concat(
                    request.Tickets.Select(t => t.ArrivalRouteAddressGuid));

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e =>
                        routeAddressGuids.Contains(e.Guid),
                    1, routeAddressGuids.Count(), cancellationToken))
                .Items;


            foreach (var t in request.Tickets)
            {
                var departureRouteAddress = routeAddresses.First(
                    ra => ra.Guid == t.DepartureRouteAddressGuid);
                var arrivalRouteAddress = routeAddresses.First(
                    ra => ra.Guid == t.ArrivalRouteAddressGuid);

                var ve = vehicleEnrollments.First(
                    e => e.Guid == t.VehicleEnrollmentGuid);

                if (departureRouteAddress.RouteId != ve.RouteId ||
                    arrivalRouteAddress.RouteId != ve.RouteId)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }

                if (departureRouteAddress.Order > arrivalRouteAddress.Order)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }
            }
        }

        // Check availability of free places.
        {
            // Get all tickets for vehicle enrollments requested in ticket group.
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var unavailableTicketStatuses = new TicketStatus[]
            {
                TicketStatus.Reserved,
                TicketStatus.Purchased
            };

            var ticketGroupTickets = (await _unitOfWork.TicketRepository
                .GetPageAsync(
                    e =>
                        vehicleEnrollmentGuids.Contains(e.VehicleEnrollment.Guid) &&
                        unavailableTicketStatuses.Contains(e.TicketGroup.Status),
                    1, int.MaxValue, cancellationToken))
                .Items;


            // Get all vehicle enrollments requested in ticket group
            // together with vehicles.
            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    e => e.Vehicle,
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            // Get all route addresses of vehicle enrollments
            // requested in ticket group.
            var routeIds = vehicleEnrollments.Select(e => e.RouteId);

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e => routeIds.Contains(e.RouteId),
                    1, int.MaxValue, cancellationToken))
                .Items;

            // For each ticket in request.
            foreach (var requestTicket in request.Tickets)
            {
                // Get vehicle enrollment of requested ticket.
                var requestVehicleEnrollment = vehicleEnrollments.First(e =>
                    e.Guid == requestTicket.VehicleEnrollmentGuid);

                // Get bought tickets of vehicle enrollment of requested ticket.
                var tickets = ticketGroupTickets.Where(t =>
                    t.VehicleEnrollmentId == requestVehicleEnrollment.Id);

                // Get route addresses of vehicle enrollment.
                var ticketRouteAddresses = routeAddresses
                    .Where(e => e.RouteId == requestVehicleEnrollment.RouteId)
                    .OrderBy(e => e.Order);


                // Count available capacity.

                // Get total capacity in requested vehicle.
                int totalCapacity;
                var vehicle = vehicleEnrollments.First(e =>
                        e.Guid == requestTicket.VehicleEnrollmentGuid)
                    .Vehicle;
                if (vehicle.VehicleType.Equals(VehicleType.Bus))
                {
                    totalCapacity = ((Bus)vehicle).Capacity;
                }
                else if (vehicle.VehicleType.Equals(VehicleType.Aircraft))
                {
                    totalCapacity = ((Aircraft)vehicle).Capacity;
                }
                else if (vehicle.VehicleType.Equals(VehicleType.Train))
                {
                    totalCapacity = ((Train)vehicle).Capacity;
                }
                else
                {
                    throw new NotImplementedException();
                }

                int takenCapacity = 0;

                // For each bought ticket.
                foreach (var ticket in tickets)
                {
                    // Get departure and arrival route address
                    // of requested ticket.
                    var requestDepartureRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Guid == requestTicket.DepartureRouteAddressGuid);
                    var requestArrivalRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Guid == requestTicket.ArrivalRouteAddressGuid);

                    // Get departure and arrival route address
                    // of bought ticket.
                    var departureRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Id == ticket.DepartureRouteAddressId);
                    var arrivalRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Id == ticket.ArrivalRouteAddressId);


                    // Count taken capacity in requested vehicle
                    // accounting for requested ticket
                    // departure and arrival route addresses.
                    // The algorithm is the same as vehicle enrollment
                    // time overlap check.
                    if ((requestDepartureRouteAddress.Order >=
                            departureRouteAddress.Order &&
                        requestDepartureRouteAddress.Order <
                            arrivalRouteAddress.Order) ||
                        (requestArrivalRouteAddress.Order <=
                            arrivalRouteAddress.Order &&
                        requestArrivalRouteAddress.Order >
                            departureRouteAddress.Order) ||
                        (requestDepartureRouteAddress.Order <=
                            departureRouteAddress.Order &&
                        requestArrivalRouteAddress.Order >=
                            arrivalRouteAddress.Order))
                    {
                        takenCapacity++;
                    }
                }

                var availableCapacity = totalCapacity - takenCapacity;

                if (availableCapacity <= 0)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }
            }
        }

        // Calculate travel time and cost.

        var ticketsDetails = new List<(short order, DateTimeOffset departureTime,
            DateTimeOffset arrivalTime, decimal cost, Currency currency)>();

        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            var routeAddressGuids =
                request.Tickets.Select(t => t.DepartureRouteAddressGuid).Concat(
                    request.Tickets.Select(t => t.ArrivalRouteAddressGuid));

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e =>
                        routeAddressGuids.Contains(e.Guid),
                    1, routeAddressGuids.Count(), cancellationToken))
                .Items;


            var vehicleEnrollmentIds = vehicleEnrollments.Select(ve => ve.Id);

            var allRouteAddressDetails = (await _unitOfWork
                .RouteAddressDetailRepository.GetPageAsync(
                    e => vehicleEnrollmentIds.Contains(e.VehicleEnrollmentId),
                    e => e.RouteAddress,
                    1, int.MaxValue, cancellationToken))
                .Items;


            foreach (var t in request.Tickets.OrderBy(t => t.Order))
            {
                var ve = vehicleEnrollments.First(
                    e => e.Guid == t.VehicleEnrollmentGuid);

                var departureRouteAddressId = routeAddresses.First(
                        ra => ra.Guid == t.DepartureRouteAddressGuid)
                    .Id;
                var arrivalRouteAddressId = routeAddresses.First(
                        ra => ra.Guid == t.ArrivalRouteAddressGuid)
                    .Id;

                var verad = allRouteAddressDetails
                    .Where(arad => arad.VehicleEnrollmentId == ve.Id)
                    .OrderBy(rad => rad.RouteAddress.Order)
                    .TakeWhile(rad => rad.Id != arrivalRouteAddressId);


                // TODO: This counts departure address stop time which is
                // not wrong but may be not desired.
                var timeToDeparture = verad
                    .TakeWhile(rad => rad.Id != departureRouteAddressId)
                    .Aggregate(TimeSpan.Zero, (sum, next) =>
                        sum + next.TimeToNextAddress + next.CurrentAddressStopTime);

                var departureTime = ve.DepartureTime.Add(timeToDeparture);


                var timeToArrival = verad.Aggregate(TimeSpan.Zero, (sum, next) =>
                    sum + next.TimeToNextAddress + next.CurrentAddressStopTime);

                var arrivalTime = ve.DepartureTime.Add(timeToArrival);


                var costToDeparture = verad
                    .TakeWhile(rad => rad.Id != departureRouteAddressId)
                    .Aggregate((decimal)0, (sum, next) =>
                        sum + next.CostToNextAddress);

                var costToArrival = verad
                    .Aggregate((decimal)0, (sum, next) =>
                        sum + next.CostToNextAddress);

                var cost = costToArrival - costToDeparture;


                ticketsDetails.Add(
                    (t.Order, departureTime, arrivalTime, cost, ve.Currency));
            }
        }

        // Check whether there are overlaps in ticket departure/arrival times.
        {
            for (int i = 1; i < ticketsDetails.Count; i++)
            {
                var previousTd = ticketsDetails[i - 1];
                var currentTd = ticketsDetails[i];

                if (previousTd.arrivalTime >= currentTd.departureTime)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }
            }
        }

        // Create entity and insert into a datastore.

        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            var routeAddressGuids =
                request.Tickets.Select(t => t.DepartureRouteAddressGuid).Concat(
                    request.Tickets.Select(t => t.ArrivalRouteAddressGuid));

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e => routeAddressGuids.Contains(e.Guid),
                    e => e.Address.City.Region.Country,
                    1, routeAddressGuids.Count(), cancellationToken))
                .Items;


            var travelTime =
                ticketsDetails.OrderBy(td => td.order).Last().arrivalTime -
                ticketsDetails.OrderBy(td => td.order).First().departureTime;

            var entity = new TicketGroup()
            {
                PassangerFirstName = request.PassangerFirstName,
                PassangerLastName = request.PassangerLastName,
                PassangerPatronymic = request.PassangerPatronymic,
                PassangerSex = request.PassangerSex,
                PassangerBirthDate = request.PassangerBirthDate,
                PurchaseTime = DateTimeOffset.UtcNow,
                Status = TicketStatus.Reserved,
                TravelTime = travelTime,
                Tickets = request.Tickets.Select(
                    t =>
                    {
                        var ve = vehicleEnrollments.First(
                            ve => ve.Guid == t.VehicleEnrollmentGuid);


                        var departureRouteAddress = routeAddresses.First(
                                ra => ra.Guid == t.DepartureRouteAddressGuid);
                        var arrivalRouteAddress = routeAddresses.First(
                                ra => ra.Guid == t.ArrivalRouteAddressGuid);


                        var detail = ticketsDetails
                            .SingleOrDefault(td => td.order == t.Order);

                        var currency = Currency.UAH;
                        var cost = _currencyConverterService
                            .ConvertAsync(
                                detail.cost, detail.currency, currency,
                                cancellationToken).Result;

                        return new Ticket()
                        {
                            DepartureRouteAddressId = departureRouteAddress.Id,
                            DepartureRouteAddress = departureRouteAddress,
                            ArrivalRouteAddressId = arrivalRouteAddress.Id,
                            ArrivalRouteAddress = arrivalRouteAddress,
                            Order = t.Order,
                            Cost = cost,
                            Currency = currency,
                            VehicleEnrollmentId = ve.Id
                        };
                    })
                    .ToArray()
            };

            entity = await _unitOfWork.TicketGroupRepository.AddOneAsync(
                entity, cancellationToken);

            await _unitOfWork.SaveAsync(cancellationToken);
            _unitOfWork.Dispose();


            var amount = entity.Tickets.Sum(e => e.Cost);
            var guid = entity.Guid;
            var validity = TimeSpan.FromMinutes(10);
            var resultPath = request.ResultPath;
            var callbackPath = "/payments/liqPay/ticket/callback";

            var paymentLink = _liqPayPaymentService
                .GetPaymentLink(
                    amount, Currency.UAH, guid.ToString(), validity,
                    _localizer["PaymentProcessing.TicketPaymentDescription"],
                    resultPath, callbackPath);

            return new PaymentLinkDto() { PaymentLink = paymentLink };
        }
    }
}
