using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using cuqmbr.TravelGuide.Domain.Entities;
using AutoMapper;
using QuikGraph;
using QuikGraph.Algorithms;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Enums;

namespace cuqmbr.TravelGuide.Application
    .VehicleEnrollmentSearch.Queries.SearchShortest;

// TODO: Add configurable time between transfers.
// TODO: Refactor DTO creation code to use mapper as much as possible.
public class SearchShortestQueryHandler :
    IRequestHandler<SearchShortestQuery, VehicleEnrollmentSearchDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly CurrencyConverterService _currencyConverterService;

    private readonly SessionTimeZoneService _sessionTimeZoneService;

    public SearchShortestQueryHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        SessionCurrencyService sessionCurrencyService,
        CurrencyConverterService currencyConverterService,
        SessionTimeZoneService sessionTimeZoneService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _sessionCurrencyService = sessionCurrencyService;
        _currencyConverterService = currencyConverterService;
        _sessionTimeZoneService = sessionTimeZoneService;
    }

    public async Task<VehicleEnrollmentSearchDto> Handle(
        SearchShortestQuery request,
        CancellationToken cancellationToken)
    {
        // Get related data

        var zeroTime = TimeOnly.FromTimeSpan(TimeSpan.Zero);
        var departureDate =
            new DateTimeOffset(request.DepartureDate, zeroTime, TimeSpan.Zero);

        var range = TimeSpan.FromDays(3);
        var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
            .GetPageAsync(
                e => 
                    e.DepartureTime >= departureDate.Subtract(range) &&
                    e.DepartureTime <= departureDate.Add(range) &&
                    request.VehicleTypes.Contains(e.Vehicle.VehicleType),
                e => e.Route,
                1, int.MaxValue, cancellationToken))
            .Items;

        if (vehicleEnrollments.Count == 0)
        {
            throw new NotFoundException();
        }

        var vehicleEnrollmentIds = vehicleEnrollments.Select(e => e.Id);
        var routeAddressDetails = (await _unitOfWork.RouteAddressDetailRepository
            .GetPageAsync(
                e => vehicleEnrollmentIds.Contains(e.VehicleEnrollmentId),
                e => e.RouteAddress.Address.City.Region.Country,
                1, int.MaxValue, cancellationToken))
            .Items;

        var vehicles = (await _unitOfWork.VehicleRepository
            .GetPageAsync(
                e => e.Enrollments.All(e => vehicleEnrollmentIds.Contains(e.Id)),
                1, int.MaxValue, cancellationToken))
            .Items;

        var companyIds = vehicles.Select(e => e.CompanyId);
        var companies = (await _unitOfWork.CompanyRepository
            .GetPageAsync(
                e => companyIds.Contains(e.Id),
                1, int.MaxValue, cancellationToken))
            .Items;


        // Hydrate vehicle enrollments with:
        //     - route address details;
        //     - vehicle info;
        //     - comapny info.

        foreach (var vehicleEnrollment in vehicleEnrollments)
        {
            vehicleEnrollment.RouteAddressDetails = routeAddressDetails
                .Where(e => e.VehicleEnrollmentId == vehicleEnrollment.Id)
                .OrderBy(e => e.RouteAddress.Order)
                .ToArray();

            vehicleEnrollment.Vehicle = vehicles
                .Single(e => e.Id == vehicleEnrollment.VehicleId);

            vehicleEnrollment.Vehicle.Company = companies
                .Single(e => e.Id == vehicleEnrollment.Vehicle.CompanyId);
        }


        // Creat and fill graph data structure

        var graph = new AdjacencyGraph<
            Address, TaggedEdge<Address, RouteAddressDetail>>();

        foreach (var vehicleEnrollment in vehicleEnrollments)
        {
            var vehicleEnrollmentRouteAddressDetails = routeAddressDetails
                .Where(e => e.VehicleEnrollmentId == vehicleEnrollment.Id)
                .OrderBy(e => e.RouteAddress.Order);

            for (int i = 1; i < vehicleEnrollmentRouteAddressDetails.Count(); i++)
            {
                var sourceRouteAddressDetail =
                    vehicleEnrollmentRouteAddressDetails.ElementAt(i-1);
                var targetRouteAddressDetail =
                    vehicleEnrollmentRouteAddressDetails.ElementAt(i);

                var sourceAddress =
                    sourceRouteAddressDetail.RouteAddress.Address;
                var targetAddress =
                    targetRouteAddressDetail.RouteAddress.Address;

                var weight = sourceRouteAddressDetail.CostToNextAddress;
                
                graph.AddVerticesAndEdge(
                    new TaggedEdge<Address, RouteAddressDetail>(
                        sourceAddress, targetAddress, sourceRouteAddressDetail));
            }
        }


        // Find paths

        var departureAddress = routeAddressDetails
            .First(e => e.RouteAddress.Address.Guid == request.DepartureAddressGuid)
            .RouteAddress.Address;
        var arrivalAddress = routeAddressDetails
            .First(e => e.RouteAddress.Address.Guid == request.ArrivalAddressGuid)
            .RouteAddress.Address;


        Func<TaggedEdge<Address, RouteAddressDetail>, double> weightByCost =
            edge => (double)edge.Tag.CostToNextAddress;

        Func<TaggedEdge<Address, RouteAddressDetail>, double> weightByTime =
            edge =>
                edge.Tag.TimeToNextAddress.Ticks +
                edge.Tag.CurrentAddressStopTime.Ticks;

        Func<TaggedEdge<Address, RouteAddressDetail>, double> edgeWeight =
            _ => 0;

        if (request.ShortestByCost && request.ShortestByTime)
        {
            edgeWeight = edge => weightByCost(edge) + weightByTime(edge);
        }
        else if (request.ShortestByCost)
        {
            edgeWeight = edge => weightByCost(edge);
        }
        else if (request.ShortestByTime)
        {
            edgeWeight = edge => weightByTime(edge);
        }


        var tryGetPaths = graph.ShortestPathsDijkstra(edgeWeight, departureAddress);


        // Create and hydrate a DTO object

        var vehicleEnrollmentDtos =
            new List<VehicleEnrollmentSearchVehicleEnrollmentDto>();

        var totalTravelTime = TimeSpan.Zero;
        var totalCost = (decimal)0;

        if (tryGetPaths(arrivalAddress, out var path))
        {
            var firstRouteAddressId = path.First().Tag.RouteAddressId;
            long lastRouteAddressId;
            Guid lastRouteAddressGuid;

            decimal vehicleEnrollmentCost;
            Currency vehicleEnrollmentCurrency;

            decimal costToNextAddress;

            var addressDtos = new List<VehicleEnrollmentSearchAddressDto>();

            var addressOrder = (short)1;
            var enrollmentOrder = (short)1;

            Address source;
            Address target;
            RouteAddressDetail tag;
            RouteAddressDetail nextTag;

            for (int i = 0; i < path.Count() - 1; i++)
            {
                source = path.Select(e => e.Source).ElementAt(i);
                tag = path.Select(e => e.Tag).ElementAt(i);
                nextTag = path.Select(e => e.Tag).ElementAt(i+1);


                costToNextAddress = await _currencyConverterService
                    .ConvertAsync(tag.CostToNextAddress,
                        tag.VehicleEnrollment.Currency,
                        _sessionCurrencyService.Currency, cancellationToken);

                addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
                {
                    Uuid = source.Guid,
                    Name = source.Name,
                    Longitude = source.Latitude,
                    Latitude = source.Longitude,
                    CountryUuid = source.City.Region.Country.Guid,
                    CountryName = source.City.Region.Country.Name,
                    RegionUuid = source.City.Region.Guid,
                    RegionName = source.City.Region.Name,
                    CityUuid = source.City.Guid,
                    CityName = source.City.Name,
                    TimeToNextAddress = tag.TimeToNextAddress,
                    CostToNextAddress = costToNextAddress,
                    CurrentAddressStopTime = tag.CurrentAddressStopTime,
                    Order = addressOrder,
                    RouteAddressUuid = tag.RouteAddress.Guid
                });

                addressOrder++;


                // First address after transfer
                if (nextTag.VehicleEnrollmentId != tag.VehicleEnrollmentId)
                {
                    target = path.Select(e => e.Target).ElementAt(i);

                    lastRouteAddressGuid = vehicleEnrollments
                        .Single(e => e.Id == tag.VehicleEnrollmentId)
                        .RouteAddressDetails
                        .Select(e => e.RouteAddress)
                        .OrderBy(e => e.Order)
                        .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                        .Take(2)
                        .ElementAt(1)
                        .Guid;

                    addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
                    {
                        Uuid = target.Guid,
                        Name = target.Name,
                        Longitude = target.Latitude,
                        Latitude = target.Longitude,
                        CountryUuid = target.City.Region.Country.Guid,
                        CountryName = target.City.Region.Country.Name,
                        RegionUuid = target.City.Region.Guid,
                        RegionName = target.City.Region.Name,
                        CityUuid = target.City.Guid,
                        CityName = target.City.Name,
                        TimeToNextAddress = TimeSpan.Zero,
                        CostToNextAddress = 0,
                        CurrentAddressStopTime = TimeSpan.Zero,
                        Order = addressOrder,
                        RouteAddressUuid = lastRouteAddressGuid
                    });

                    lastRouteAddressId = vehicleEnrollments
                        .Single(e => e.Id == tag.VehicleEnrollmentId)
                        .RouteAddressDetails
                        .Select(e => e.RouteAddress)
                        .OrderBy(e => e.Order)
                        .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                        .Take(2)
                        .ElementAt(1)
                        .Id;

                    vehicleEnrollmentCost = await _currencyConverterService
                        .ConvertAsync(
                            tag.VehicleEnrollment
                                .GetCost(firstRouteAddressId,
                                    lastRouteAddressId),
                            tag.VehicleEnrollment.Currency,
                            _sessionCurrencyService.Currency,
                            cancellationToken);

                    vehicleEnrollmentCurrency =
                        _sessionCurrencyService.Currency.Equals(Currency.Default) ?
                        tag.VehicleEnrollment.Currency :
                        _sessionCurrencyService.Currency;

                    vehicleEnrollmentDtos.Add(
                        new VehicleEnrollmentSearchVehicleEnrollmentDto()
                        {
                            DepartureTime = tag.VehicleEnrollment
                                .GetDepartureTime(firstRouteAddressId)
                                .ToOffset(_sessionTimeZoneService.TimeZone.BaseUtcOffset),
                            ArrivalTime = tag.VehicleEnrollment
                                .GetArrivalTime(lastRouteAddressId)
                                .ToOffset(_sessionTimeZoneService.TimeZone.BaseUtcOffset),
                            TravelTime = tag.VehicleEnrollment
                                .GetTravelTime(firstRouteAddressId,
                                    lastRouteAddressId),
                            TimeMoving = tag.VehicleEnrollment
                                .GetTimeMoving(firstRouteAddressId,
                                    lastRouteAddressId),
                            TimeInStops = tag.VehicleEnrollment
                                .GetTimeInStops(firstRouteAddressId,
                                    lastRouteAddressId),
                            NumberOfStops = tag.VehicleEnrollment
                                .GetNumberOfStops(firstRouteAddressId,
                                    lastRouteAddressId),
                            Currency = vehicleEnrollmentCurrency.Name,
                            Cost = vehicleEnrollmentCost,
                            VehicleType = source.VehicleType.Name,
                            Uuid = tag.VehicleEnrollment.Guid,
                            Order = enrollmentOrder,
                            Addresses = addressDtos,
                            Company = _mapper
                                .Map<VehicleEnrollmentSearchCompanyDto>(
                                    tag.VehicleEnrollment.Vehicle.Company),
                            Vehicle = _mapper
                                .Map<VehicleEnrollmentSearchVehicleDto>(
                                    tag.VehicleEnrollment.Vehicle)
                        });

                    firstRouteAddressId = nextTag.RouteAddressId;

                    addressDtos = new List<VehicleEnrollmentSearchAddressDto>();
                    addressOrder = (short)1;
                    enrollmentOrder++;
                }
            }

            source = path.Select(e => e.Source).Last();
            target = path.Select(e => e.Target).Last();
            tag = path.Select(e => e.Tag).Last();
            nextTag = path.Select(e => e.Tag).Last();


            costToNextAddress = await _currencyConverterService
                .ConvertAsync(tag.CostToNextAddress,
                    tag.VehicleEnrollment.Currency,
                    _sessionCurrencyService.Currency, cancellationToken);

            addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
            {
                Uuid = source.Guid,
                Name = source.Name,
                Longitude = source.Latitude,
                Latitude = source.Longitude,
                CountryUuid = source.City.Region.Country.Guid,
                CountryName = source.City.Region.Country.Name,
                RegionUuid = source.City.Region.Guid,
                RegionName = source.City.Region.Name,
                CityUuid = source.City.Guid,
                CityName = source.City.Name,
                TimeToNextAddress = tag.TimeToNextAddress,
                CostToNextAddress = costToNextAddress,
                CurrentAddressStopTime = tag.CurrentAddressStopTime,
                Order = addressOrder,
                RouteAddressUuid = tag.RouteAddress.Guid
            });

            addressOrder++;

            lastRouteAddressGuid = vehicleEnrollments
                .Single(e => e.Id == tag.VehicleEnrollmentId)
                .RouteAddressDetails
                .Select(e => e.RouteAddress)
                .OrderBy(e => e.Order)
                .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                .Take(2)
                .ElementAt(1)
                .Guid;

            addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
            {
                Uuid = target.Guid,
                Name = target.Name,
                Longitude = target.Latitude,
                Latitude = target.Longitude,
                CountryUuid = target.City.Region.Country.Guid,
                CountryName = target.City.Region.Country.Name,
                RegionUuid = target.City.Region.Guid,
                RegionName = target.City.Region.Name,
                CityUuid = target.City.Guid,
                CityName = target.City.Name,
                TimeToNextAddress = TimeSpan.Zero,
                CostToNextAddress = 0,
                CurrentAddressStopTime = TimeSpan.Zero,
                Order = addressOrder,
                RouteAddressUuid = lastRouteAddressGuid
            });

            lastRouteAddressId = vehicleEnrollments
                .Single(e => e.Id == tag.VehicleEnrollmentId)
                .RouteAddressDetails
                .Select(e => e.RouteAddress)
                .OrderBy(e => e.Order)
                .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                .Take(2)
                .ElementAt(1)
                .Id;

            vehicleEnrollmentCost = await _currencyConverterService
                .ConvertAsync(
                    tag.VehicleEnrollment
                        .GetCost(firstRouteAddressId,
                            lastRouteAddressId),
                    tag.VehicleEnrollment.Currency,
                    _sessionCurrencyService.Currency,
                    cancellationToken);

            vehicleEnrollmentCurrency =
                _sessionCurrencyService.Currency.Equals(Currency.Default) ?
                tag.VehicleEnrollment.Currency :
                _sessionCurrencyService.Currency;

            vehicleEnrollmentDtos.Add(
                new VehicleEnrollmentSearchVehicleEnrollmentDto()
                {
                    DepartureTime = tag.VehicleEnrollment
                        .GetDepartureTime(firstRouteAddressId)
                        .ToOffset(_sessionTimeZoneService.TimeZone.BaseUtcOffset),
                    ArrivalTime = tag.VehicleEnrollment
                        .GetArrivalTime(lastRouteAddressId)
                        .ToOffset(_sessionTimeZoneService.TimeZone.BaseUtcOffset),
                    TravelTime = tag.VehicleEnrollment
                        .GetTravelTime(firstRouteAddressId,
                            lastRouteAddressId),
                    TimeMoving = tag.VehicleEnrollment
                        .GetTimeMoving(firstRouteAddressId,
                            lastRouteAddressId),
                    TimeInStops = tag.VehicleEnrollment
                        .GetTimeInStops(firstRouteAddressId,
                            lastRouteAddressId),
                    NumberOfStops = tag.VehicleEnrollment
                        .GetNumberOfStops(firstRouteAddressId,
                            lastRouteAddressId),
                    Currency = vehicleEnrollmentCurrency.Name,
                    Cost = vehicleEnrollmentCost,
                    VehicleType = source.VehicleType.Name,
                    Uuid = tag.VehicleEnrollment.Guid,
                    Order = enrollmentOrder,
                    Addresses = addressDtos,
                    Company = _mapper.Map<VehicleEnrollmentSearchCompanyDto>(
                        tag.VehicleEnrollment.Vehicle.Company),
                    Vehicle = _mapper.Map<VehicleEnrollmentSearchVehicleDto>(
                        tag.VehicleEnrollment.Vehicle)
                });
        }
        else
        {
            throw new NotFoundException();
        }


        var departureTime = vehicleEnrollmentDtos
            .OrderBy(e => e.Order).First().DepartureTime;
        var arrivalTime = vehicleEnrollmentDtos
            .OrderBy(e => e.Order).Last().ArrivalTime;
        var timeInStops = vehicleEnrollmentDtos
            .Aggregate(TimeSpan.Zero, (sum, next) => sum += next.TimeInStops);
        var numberOfTransfers = vehicleEnrollmentDtos.Count() - 1;
        var cost = vehicleEnrollmentDtos
            .Aggregate((decimal)0, (sum, next) => sum += next.Cost);

        return new VehicleEnrollmentSearchDto()
        {
            DepartureTime = departureTime
                .ToOffset(_sessionTimeZoneService.TimeZone.BaseUtcOffset),
            ArrivalTime = arrivalTime
                .ToOffset(_sessionTimeZoneService.TimeZone.BaseUtcOffset),
            TravelTime = arrivalTime - departureTime,
            TimeInStops = timeInStops,
            NumberOfTransfers = numberOfTransfers,
            Currency = _sessionCurrencyService.Currency.Name,
            Cost = cost,
            Enrollments = vehicleEnrollmentDtos
        };
    }
}
