using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.Extensions;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Enums;

namespace cuqmbr.TravelGuide.Application.VehicleEnrollments
    .Queries.GetVehicleEnrollmentsPage;

public class GetVehicleEnrollmentsPageQueryHandler :
    IRequestHandler<GetVehicleEnrollmentsPageQuery, PaginatedList<VehicleEnrollmentDto>>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly CurrencyConverterService _currencyConverterService;

    public GetVehicleEnrollmentsPageQueryHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        SessionCurrencyService sessionCurrencyService,
        CurrencyConverterService currencyConverterService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _sessionCurrencyService = sessionCurrencyService;
        _currencyConverterService = currencyConverterService;
    }

    public async Task<PaginatedList<VehicleEnrollmentDto>> Handle(
        GetVehicleEnrollmentsPageQuery request,
        CancellationToken cancellationToken)
    {
        // TODO: Add search functionality or remove it
        var pagedList = await _unitOfWork.VehicleEnrollmentRepository
            .GetPageAsync(
                e =>
                    // (e.Name.ToLower().Contains(request.Search.ToLower()) ||
                    // e.City.Region.Country.Name.ToLower().Contains(request.Search.ToLower())) &&
                    (request.RouteGuid != null
                        ? e.Route.Guid == request.RouteGuid
                        : true) &&
                    (request.VehicleGuid != null
                        ? e.Vehicle.Guid >= request.VehicleGuid
                        : true) &&
                    (request.NumberOfAddressesGreaterThanOrEqual != null
                        ? 
                            e.RouteAddressDetails.Count() >=
                            request.NumberOfAddressesGreaterThanOrEqual
                        : true) &&
                    (request.NumberOfAddressesLessThanOrEqual != null
                        ? 
                            e.RouteAddressDetails.Count() <=
                            request.NumberOfAddressesLessThanOrEqual
                        : true) &&
                    (request.DepartureTimeGreaterThanOrEqual != null
                        ? e.DepartureTime >= request.DepartureTimeGreaterThanOrEqual
                        : true) &&
                    (request.DepartureTimeLessThanOrEqual != null
                        ? e.DepartureTime <= request.DepartureTimeLessThanOrEqual
                        : true) &&
                    (request.ArrivalTimeGreaterThanOrEqual != null
                        ? 
                            e.DepartureTime.AddSeconds(e.RouteAddressDetails
                                .Sum(rad => rad.TimeToNextAddress.TotalSeconds +
                                    rad.CurrentAddressStopTime.TotalSeconds)) >=
                            request.ArrivalTimeGreaterThanOrEqual
                        : true) &&
                    (request.ArrivalTimeLessThanOrEqual != null
                        ? 
                            e.DepartureTime.AddSeconds(e.RouteAddressDetails
                                .Sum(rad => rad.TimeToNextAddress.TotalSeconds +
                                    rad.CurrentAddressStopTime.TotalSeconds)) <=
                            request.ArrivalTimeLessThanOrEqual
                        : true) &&
                    (request.TravelTimeGreaterThanOrEqual != null
                        ? 
                            e.RouteAddressDetails
                                .Sum(rad => rad.TimeToNextAddress.TotalSeconds) >=
                            request.TravelTimeGreaterThanOrEqual.Value.TotalSeconds
                        : true) &&
                    (request.TravelTimeLessThanOrEqual != null
                        ? 
                            e.RouteAddressDetails
                                .Sum(rad => rad.TimeToNextAddress.TotalSeconds) <=
                            request.TravelTimeLessThanOrEqual.Value.TotalSeconds
                        : true) &&
                    (request.TimeMovingGreaterThanOrEqual != null
                        ? 
                            e.RouteAddressDetails
                                .Sum(rad => rad.TimeToNextAddress.TotalSeconds) >=
                            request.TimeMovingGreaterThanOrEqual.Value.TotalSeconds
                        : true) &&
                    (request.TimeMovingLessThanOrEqual != null
                        ? 
                            e.RouteAddressDetails
                                .Sum(rad => rad.TimeToNextAddress.TotalSeconds) <=
                            request.TimeMovingLessThanOrEqual.Value.TotalSeconds
                        : true) &&
                    (request.TimeInStopsGreaterThanOrEqual != null
                        ? 
                            e.RouteAddressDetails
                                .Sum(rad => rad.CurrentAddressStopTime.TotalSeconds) >=
                            request.TimeInStopsGreaterThanOrEqual.Value.TotalSeconds
                        : true) &&
                    (request.TimeInStopsLessThanOrEqual != null
                        ? 
                            e.RouteAddressDetails
                                .Sum(rad => rad.CurrentAddressStopTime.TotalSeconds) <=
                            request.TimeInStopsLessThanOrEqual.Value.TotalSeconds
                        : true) &&
                    (request.CostGreaterThanOrEqual != null
                        ? e.RouteAddressDetails.Sum(rad => rad.CostToNextAddress) >=
                            request.CostGreaterThanOrEqual
                        : true) &&
                    (request.CostLessThanOrEqual != null
                        ? e.RouteAddressDetails.Sum(rad => rad.CostToNextAddress) <=
                            request.CostLessThanOrEqual
                        : true) &&
                    (request.Currency != null
                        ? e.Currency == request.Currency
                        : true) &&
                    (request.EmployeeGuids != null
                        ? e.VehicleEnrollmentEmployees.Any(e => request.EmployeeGuids.Contains(e.Employee.Guid))
                        : true),
                e => e.RouteAddressDetails,
                request.PageNumber, request.PageSize, cancellationToken);

        var vehicleEnrollments = pagedList.Items;

        var vehicleIds = vehicleEnrollments.Select(e => e.VehicleId);
        var vehicles = (await _unitOfWork.VehicleRepository
            .GetPageAsync(
                e => vehicleIds.Contains(e.Id),
                1, vehicleIds.Count(), cancellationToken))
            .Items;

        var routeIds = vehicleEnrollments.Select(e => e.RouteId);
        var routes = (await _unitOfWork.RouteRepository
            .GetPageAsync(
                e => routeIds.Contains(e.Id),
                1, routeIds.Count(), cancellationToken))
            .Items;


        // Hydrate vehicle enrollment with address,
        // vehicle, route and employee information.

        var routeAddressIds = vehicleEnrollments
            .SelectMany(ve => ve.RouteAddressDetails)
            .Select(rad => rad.RouteAddressId);

        var routeAddresses = (await _unitOfWork.RouteAddressRepository
            .GetPageAsync(
                e =>
                    routeAddressIds.Contains(e.Id),
                e => e.Address.City.Region.Country,
                1, routeAddressIds.Count(), cancellationToken))
            .Items;

        var vehicleEnrollmentIds = vehicleEnrollments
            .Select(ve => ve.Id);

        var vehicleEnrollmentEmployees = 
            (await _unitOfWork.VehicleEnrollmentEmployeeRepository
                .GetPageAsync(
                    e => vehicleEnrollmentIds.Contains(e.VehicleEnrollmentId),
                    e => e.Employee,
                    1, int.MaxValue, cancellationToken))
                .Items;

        foreach (var vehicleEnrollment in vehicleEnrollments)
        {
            foreach (var routeAddressDetail in
                    vehicleEnrollment.RouteAddressDetails)
            {
                routeAddressDetail.RouteAddress = routeAddresses
                    .First(ra => ra.Id == routeAddressDetail.RouteAddressId);
            }

            vehicleEnrollment.Route = routes
                .Single(e => e.Id == vehicleEnrollment.RouteId);

            vehicleEnrollment.Vehicle = vehicles
                .Single(e => e.Id == vehicleEnrollment.VehicleId);

            vehicleEnrollment.VehicleEnrollmentEmployees =
                vehicleEnrollmentEmployees
                    .Where(vee => vee.VehicleEnrollmentId == vehicleEnrollment.Id)
                    .ToArray();
        }


        // Convert currency

        // TODO: Replace with AutoMapper Resolver

        if (!_sessionCurrencyService.Currency.Equals(Currency.Default))
        {
            foreach (var ve in vehicleEnrollments)
            {
                foreach (var rad in ve.RouteAddressDetails)
                {
                    rad.CostToNextAddress = await _currencyConverterService
                        .ConvertAsync(rad.CostToNextAddress, ve.Currency,
                            _sessionCurrencyService.Currency, cancellationToken);
                }
                ve.Currency = _sessionCurrencyService.Currency;
            }
        }


        var vehicleEnrollmentsDto = _mapper
            .Map<List<VehicleEnrollmentDto>>(vehicleEnrollments)
            .AsQueryable();

        vehicleEnrollmentsDto = QueryableExtension<VehicleEnrollmentDto>
            .ApplySort(vehicleEnrollmentsDto, request.Sort);

        _unitOfWork.Dispose();

        return new PaginatedList<VehicleEnrollmentDto>(
            vehicleEnrollmentsDto.ToList(),
            pagedList.TotalCount, request.PageNumber,
            request.PageSize);
    }
}
