using System.Globalization;
using System.Text;
using AspNetCore.Localizer.Json.Commons;
using AspNetCore.Localizer.Json.Extensions;
using cuqmbr.TravelGuide.Application;
using cuqmbr.TravelGuide.Application.Common.Behaviours;
using FluentValidation;
using MediatR;
using MediatR.Behaviors.Authorization.Extensions.DependencyInjection;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Options;
using cuqmbr.TravelGuide.Application.Common.Authorization;
using Microsoft.AspNetCore.Authentication.JwtBearer;
using Microsoft.IdentityModel.Tokens;

namespace cuqmbr.TravelGuide.Configuration.Application;

public static class Configuration
{
    public static IServiceCollection ConfigureApplication(
        this IServiceCollection services)
    {
        var configuration = services.BuildServiceProvider().GetService<
            IOptions<cuqmbr.TravelGuide.Application.ConfigurationOptions>>()
            .Value;

        return services
            .AddLocalization(configuration.Localization)
            .AddFluentValidation()
            .AddAutoMapper()
            .AddMediatR()
            .AddAuthentication(configuration.JsonWebToken);
    }

    private static IServiceCollection AddFluentValidation(
        this IServiceCollection services)
    {
        return services.AddValidatorsFromAssembly(
            typeof(cuqmbr.TravelGuide.Application.ConfigurationOptions).Assembly);
    }

    private static IServiceCollection AddAutoMapper(
        this IServiceCollection services)
    {
        return services.AddAutoMapper(
            typeof(cuqmbr.TravelGuide.Application.ConfigurationOptions).Assembly);
    }

    private static IServiceCollection AddMediatR(
        this IServiceCollection services)
    {
        services.AddMediatorAuthorization(
            typeof(cuqmbr.TravelGuide.Application.ConfigurationOptions).Assembly,
            configuration =>
                configuration.UseUnauthorizedResultHandlerStrategy(
                    new ForbiddenUnauthorizedResultHandler()));

        services.AddAuthorizersFromAssembly(
            typeof(cuqmbr.TravelGuide.Application.ConfigurationOptions).Assembly);

        services.AddMediatR(configuration =>
        {
            configuration.RegisterServicesFromAssembly(
                typeof(cuqmbr.TravelGuide.Application.ConfigurationOptions)
                    .Assembly);
            configuration.AddBehavior(
                typeof(IPipelineBehavior<,>),
                typeof(LoggingBehaviour<,>));
            configuration.AddBehavior(
                typeof(IPipelineBehavior<,>),
                typeof(ValidationBehaviour<,>));
        });

        return services;
    }

    private static IServiceCollection AddLocalization(
        this IServiceCollection services,
        LocalizationConfigurationOptions configuration)
    {
        return services.AddJsonLocalization(options =>
        {
            options.ResourcesPath = "Resources";
            options.FileEncoding = Encoding.UTF8;
            options.DefaultCulture =
                CultureInfo.GetCultureInfo(configuration.DefaultCultureName);
            options.DefaultUICulture =
                CultureInfo.GetCultureInfo(configuration.DefaultCultureName);
            options.LocalizationMode =
                AspNetCore.Localizer.Json.JsonOptions.LocalizationMode.I18n;
            options.CacheDuration = configuration.CacheDuration;
            options.AssemblyHelper = new AssemblyHelper(
                typeof(cuqmbr.TravelGuide.Application.ConfigurationOptions)
                .Assembly);
        });
    }

    private static IServiceCollection AddAuthentication(
        this IServiceCollection services,
        JsonWebTokenConfigurationOptions configuration)
    {
        services
            .AddAuthentication(options =>
            {
                options.DefaultAuthenticateScheme =
                    JwtBearerDefaults.AuthenticationScheme;
                options.DefaultChallengeScheme =
                    JwtBearerDefaults.AuthenticationScheme;
                options.DefaultScheme =
                    JwtBearerDefaults.AuthenticationScheme;
            })
            .AddJwtBearer(options =>
            {
                options.IncludeErrorDetails = true;
                options.SaveToken = true;
                options.RequireHttpsMetadata = false;
                options.TokenValidationParameters =
                    new TokenValidationParameters()
                    {
                        ValidateIssuer = false,
                        ValidateAudience = false,
                        ValidateLifetime = true,
                        ValidateIssuerSigningKey = true,
                        ValidAudience = configuration.Audience,
                        ValidIssuer = configuration.Issuer,
                        ClockSkew = TimeSpan.Zero,
                        IssuerSigningKey = new SymmetricSecurityKey(
                            Encoding.UTF8.GetBytes(
                                configuration.IssuerSigningKey))
                    };
            });

        return services;
    }
}
