using Microsoft.AspNetCore.Mvc;
using Swashbuckle.AspNetCore.Annotations;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.ViewModels;
using cuqmbr.TravelGuide.Domain.Enums;
using cuqmbr.TravelGuide.Application.Employees;
using cuqmbr.TravelGuide.Application.Employees.Commands.AddEmployee;
using cuqmbr.TravelGuide.Application.Employees.Queries.GetEmployeesPage;
using cuqmbr.TravelGuide.Application.Employees.Queries.GetEmployee;
using cuqmbr.TravelGuide.Application.Employees.Commands.UpdateEmployee;
using cuqmbr.TravelGuide.Application.Employees.Commands.DeleteEmployee;
using cuqmbr.TravelGuide.Application.Employees.ViewModels;
using cuqmbr.TravelGuide.Application.Employees.Models;

namespace cuqmbr.TravelGuide.HttpApi.Controllers;

[Route("employees")]
public class EmployeesController : ControllerBase
{
    [HttpPost]
    [SwaggerOperation("Add an employee")]
    [SwaggerResponse(
        StatusCodes.Status201Created, "Object successfuly created",
        typeof(EmployeeDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "One or more addresses was not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<ActionResult<EmployeeDto>> Add(
        [FromBody] AddEmployeeViewModel viewModel,
        CancellationToken cancellationToken)
    {
        return StatusCode(
            StatusCodes.Status201Created,
            await Mediator.Send(
                new AddEmployeeCommand()
                {
                    FirstName = viewModel.FirstName,
                    LastName = viewModel.LastName,
                    Patronymic = viewModel.Patronymic,
                    Sex = Sex.FromName(viewModel.Sex),
                    BirthDate = viewModel.BirthDate,
                    Documents = viewModel.Documents.Select(
                        e => new EmployeeDocumentModel()
                        {
                            DocumentType = DocumentType.FromName(e.DocumentType),
                            Information = e.Information
                           
                        }).ToArray(),
                    CompanyGuid = viewModel.CompanyUuid
                },
                cancellationToken));
    }

    [HttpGet]
    [SwaggerOperation("Get a list of all employees")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful",
        typeof(PaginatedList<EmployeeDto>))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<PaginatedList<EmployeeDto>> GetPage(
        [FromQuery] PageQuery pageQuery, [FromQuery] SearchQuery searchQuery,
        [FromQuery] SortQuery sortQuery, 
        [FromQuery] GetEmployeesPageFilterViewModel filterQuery,
        CancellationToken cancellationToken)
    {
        return await Mediator.Send(
            new GetEmployeesPageQuery()
            {
                PageNumber = pageQuery.PageNumber,
                PageSize = pageQuery.PageSize,
                Search = searchQuery.Search,
                Sort = sortQuery.Sort,
                Sex = Sex.FromName(filterQuery.Sex),
                BirthDateLessThanOrEqualTo = 
                    filterQuery.BirthDateLessThanOrEqualTo,
                BirthDateGreaterThanOrEqualTo =
                    filterQuery.BirthDateGreaterThanOrEqualTo,
                CompanyGuid = filterQuery.CompanyUuid
            },
            cancellationToken);
    }

    [HttpGet("{uuid:guid}")]
    [SwaggerOperation("Get an employee by uuid")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful", typeof(EmployeeDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Object not found", typeof(EmployeeDto))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<EmployeeDto> Get(
        [FromRoute] Guid uuid,
        CancellationToken cancellationToken)
    {
        return await Mediator.Send(new GetEmployeeQuery() { Guid = uuid },
            cancellationToken);
    }

    [HttpPut("{uuid:guid}")]
    [SwaggerOperation("Update an employee")]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Object already exists",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Object not found", typeof(EmployeeDto))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "One or more addresses was not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<EmployeeDto> Update(
        [FromRoute] Guid uuid,
        [FromBody] UpdateEmployeeViewModel viewModel,
        CancellationToken cancellationToken)
    {
            return await Mediator.Send(
                new UpdateEmployeeCommand()
                {
                    Guid = uuid,
                    FirstName = viewModel.FirstName,
                    LastName = viewModel.LastName,
                    Patronymic = viewModel.Patronymic,
                    Sex = Sex.FromName(viewModel.Sex),
                    BirthDate = viewModel.BirthDate,
                    Documents = viewModel.Documents.Select(
                        e => new EmployeeDocumentModel()
                        {
                            DocumentType = DocumentType.FromName(e.DocumentType),
                            Information = e.Information
                           
                        }).ToArray(),
                    CompanyGuid = viewModel.CompanyUuid
                },
                cancellationToken);
    }

    [HttpDelete("{uuid:guid}")]
    [SwaggerOperation("Delete an employee")]
    [SwaggerResponse(StatusCodes.Status204NoContent, "Request successful")]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Object not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<IActionResult> Delete(
        [FromRoute] Guid uuid,
        CancellationToken cancellationToken)
    {
        await Mediator.Send(
            new DeleteEmployeeCommand() { Guid = uuid },
            cancellationToken);
        return StatusCode(StatusCodes.Status204NoContent);
    }
}
