using Microsoft.AspNetCore.Mvc;
using Swashbuckle.AspNetCore.Annotations;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.ViewModels;
using cuqmbr.TravelGuide.Domain.Enums;
using cuqmbr.TravelGuide.Application.Identity.Roles.Queries.GetRolesPage;
using cuqmbr.TravelGuide.Application.Identity.Accounts;
using cuqmbr.TravelGuide.Application.Identity.Accounts.ViewModels;
using cuqmbr.TravelGuide.Application.Identity.Accounts.Commands.AddAccount;
// using cuqmbr.TravelGuide.Application.Identity.Commands.AddIdentity;
// using cuqmbr.TravelGuide.Application.Identity.Queries.GetIdentityPage;
// using cuqmbr.TravelGuide.Application.Identity.Queries.GetIdentity;
// using cuqmbr.TravelGuide.Application.Identity.Commands.UpdateIdentity;
// using cuqmbr.TravelGuide.Application.Identity.Commands.DeleteIdentity;
// using cuqmbr.TravelGuide.Application.Identity.ViewModels;

namespace cuqmbr.TravelGuide.HttpApi.Controllers;

[Route("identity")]
public class IdentityController : ControllerBase
{
    [HttpGet("roles")]
    [SwaggerOperation("Get a list of all roles")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful",
        typeof(PaginatedList<string>))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<PaginatedList<string>> GetRolesPage(
        [FromQuery] PageQuery pageQuery, [FromQuery] SearchQuery searchQuery,
        CancellationToken cancellationToken)
    {
        return await Mediator.Send(
            new GetRolesPageQuery()
            {
                PageNumber = pageQuery.PageNumber,
                PageSize = pageQuery.PageSize,
                Search = searchQuery.Search
            },
            cancellationToken);
    }




    [HttpPost("accounts")]
    [SwaggerOperation("Add an account")]
    [SwaggerResponse(
        StatusCodes.Status201Created, "Object successfuly created",
        typeof(AccountDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Object already exists",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Parent object not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<ActionResult<AccountDto>> Add(
        [FromBody] AddAccountViewModel viewModel,
        CancellationToken cancellationToken)
    {
        return StatusCode(
            StatusCodes.Status201Created,
            await Mediator.Send(
                new AddAccountCommand()
                {
                    Email = viewModel.Email,
                    Password = viewModel.Password,
                    Roles = viewModel.Roles
                        .Select(s => IdentityRole.FromName(s))
                        .ToArray()
                },
                cancellationToken));
    }




    // [HttpPost]
    // [SwaggerOperation("Add an identity")]
    // [SwaggerResponse(
    //     StatusCodes.Status201Created, "Object successfuly created",
    //     typeof(IdentityDto))]
    // [SwaggerResponse(
    //     StatusCodes.Status400BadRequest, "Object already exists",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status400BadRequest, "Input data validation error",
    //     typeof(HttpValidationProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status403Forbidden,
    //     "Not enough privileges to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status404NotFound, "Parent object not found",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status500InternalServerError, "Internal server error",
    //     typeof(ProblemDetails))]
    // public async Task<ActionResult<IdentityDto>> Add(
    //     [FromBody] AddIdentityViewModel viewModel,
    //     CancellationToken cancellationToken)
    // {
    //     return StatusCode(
    //         StatusCodes.Status201Created,
    //         await Mediator.Send(
    //             new AddIdentityCommand()
    //             {
    //                 Name = viewModel.Name,
    //                 Longitude = viewModel.Longitude,
    //                 Latitude = viewModel.Latitude,
    //                 VehicleType = VehicleType.FromName(viewModel.VehicleType),
    //                 CityGuid = viewModel.CityUuid
    //             },
    //             cancellationToken));
    // }
    //
    // [HttpGet("{uuid:guid}")]
    // [SwaggerOperation("Get an identity by uuid")]
    // [SwaggerResponse(
    //     StatusCodes.Status200OK, "Request successful", typeof(IdentityDto))]
    // [SwaggerResponse(
    //     StatusCodes.Status400BadRequest, "Input data validation error",
    //     typeof(HttpValidationProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status403Forbidden,
    //     "Not enough privileges to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status404NotFound, "Object not found", typeof(IdentityDto))]
    // [SwaggerResponse(
    //     StatusCodes.Status500InternalServerError, "Internal server error",
    //     typeof(ProblemDetails))]
    // public async Task<IdentityDto> Get(
    //     [FromRoute] Guid uuid,
    //     CancellationToken cancellationToken)
    // {
    //     return await Mediator.Send(new GetIdentityQuery() { Guid = uuid },
    //         cancellationToken);
    // }
    //
    // [HttpPut("{uuid:guid}")]
    // [SwaggerOperation("Update an identity")]
    // [SwaggerResponse(
    //     StatusCodes.Status200OK, "Request successful", typeof(IdentityDto))]
    // [SwaggerResponse(
    //     StatusCodes.Status400BadRequest, "Object already exists",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status400BadRequest, "Input data validation error",
    //     typeof(HttpValidationProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status403Forbidden,
    //     "Not enough privileges to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status404NotFound, "Object not found", typeof(IdentityDto))]
    // [SwaggerResponse(
    //     StatusCodes.Status404NotFound, "Parent object not found",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status500InternalServerError, "Internal server error",
    //     typeof(ProblemDetails))]
    // public async Task<IdentityDto> Update(
    //     [FromRoute] Guid uuid,
    //     [FromBody] UpdateIdentityViewModel viewModel,
    //     CancellationToken cancellationToken)
    // {
    //         return await Mediator.Send(
    //             new UpdateIdentityCommand()
    //             {
    //                 Guid = uuid,
    //                 Name = viewModel.Name,
    //                 Longitude = viewModel.Longitude,
    //                 Latitude = viewModel.Latitude,
    //                 VehicleType = VehicleType.FromName(viewModel.VehicleType),
    //                 CityGuid = viewModel.CityUuid
    //             },
    //             cancellationToken);
    // }
    //
    // [HttpDelete("{uuid:guid}")]
    // [SwaggerOperation("Delete an identity")]
    // [SwaggerResponse(StatusCodes.Status204NoContent, "Request successful")]
    // [SwaggerResponse(
    //     StatusCodes.Status400BadRequest, "Input data validation error",
    //     typeof(HttpValidationProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status403Forbidden,
    //     "Not enough privileges to perform an action",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status404NotFound, "Object not found",
    //     typeof(ProblemDetails))]
    // [SwaggerResponse(
    //     StatusCodes.Status500InternalServerError, "Internal server error",
    //     typeof(ProblemDetails))]
    // public async Task<IActionResult> Delete(
    //     [FromRoute] Guid uuid,
    //     CancellationToken cancellationToken)
    // {
    //     await Mediator.Send(
    //         new DeleteIdentityCommand() { Guid = uuid },
    //         cancellationToken);
    //     return StatusCode(StatusCodes.Status204NoContent);
    // }
}
