using Microsoft.AspNetCore.Mvc;
using Swashbuckle.AspNetCore.Annotations;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.ViewModels;
using cuqmbr.TravelGuide.Domain.Enums;
using cuqmbr.TravelGuide.Application.Trains;
using cuqmbr.TravelGuide.Application.Trains.Commands.AddTrain;
using cuqmbr.TravelGuide.Application.Trains.Queries.GetTrainsPage;
using cuqmbr.TravelGuide.Application.Trains.Queries.GetTrain;
using cuqmbr.TravelGuide.Application.Trains.Commands.UpdateTrain;
using cuqmbr.TravelGuide.Application.Trains.Commands.DeleteTrain;
using cuqmbr.TravelGuide.Application.Trains.ViewModels;

namespace cuqmbr.TravelGuide.HttpApi.Controllers;

[Route("trains")]
public class TrainsController : ControllerBase
{
    [HttpPost]
    [SwaggerOperation("Add a train")]
    [SwaggerResponse(
        StatusCodes.Status201Created, "Object successfuly created",
        typeof(TrainDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Object already exists",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Parent object not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<ActionResult<TrainDto>> Add(
        [FromBody] AddTrainViewModel viewModel,
        CancellationToken cancellationToken)
    {
        return StatusCode(
            StatusCodes.Status201Created,
            await Mediator.Send(
                new AddTrainCommand()
                {
                    Number = viewModel.Number,
                    Model = viewModel.Model,
                    Capacity = viewModel.Capacity
                },
                cancellationToken));
    }

    [HttpGet]
    [SwaggerOperation("Get a list of all trains")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful",
        typeof(PaginatedList<TrainDto>))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<PaginatedList<TrainDto>> GetPage(
        [FromQuery] PageQuery pageQuery, [FromQuery] SearchQuery searchQuery,
        [FromQuery] SortQuery sortQuery, 
        [FromQuery] GetTrainsPageFilterViewModel filterQuery,
        CancellationToken cancellationToken)
    {
        return await Mediator.Send(
            new GetTrainsPageQuery()
            {
                PageNumber = pageQuery.PageNumber,
                PageSize = pageQuery.PageSize,
                Search = searchQuery.Search,
                Sort = sortQuery.Sort,
                CapacityGreaterOrEqualThan = 
                    filterQuery.CapacityGreaterThanOrEqualTo,
                CapacityLessOrEqualThan = 
                    filterQuery.CapacityLessThanOrEqualTo
            },
            cancellationToken);
    }

    [HttpGet("{uuid:guid}")]
    [SwaggerOperation("Get a train by uuid")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful", typeof(TrainDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Object not found", typeof(TrainDto))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<TrainDto> Get(
        [FromRoute] Guid uuid,
        CancellationToken cancellationToken)
    {
        return await Mediator.Send(new GetTrainQuery() { Guid = uuid },
            cancellationToken);
    }

    [HttpPut("{uuid:guid}")]
    [SwaggerOperation("Update a train")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful", typeof(TrainDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Object already exists",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Object not found", typeof(TrainDto))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Parent object not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<TrainDto> Update(
        [FromRoute] Guid uuid,
        [FromBody] UpdateTrainViewModel viewModel,
        CancellationToken cancellationToken)
    {
            return await Mediator.Send(
                new UpdateTrainCommand()
                {
                    Guid = uuid,
                    Number = viewModel.Number,
                    Model = viewModel.Model,
                    Capacity = viewModel.Capacity
                },
                cancellationToken);
    }

    [HttpDelete("{uuid:guid}")]
    [SwaggerOperation("Delete a train")]
    [SwaggerResponse(StatusCodes.Status204NoContent, "Request successful")]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Object not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<IActionResult> Delete(
        [FromRoute] Guid uuid,
        CancellationToken cancellationToken)
    {
        await Mediator.Send(
            new DeleteTrainCommand() { Guid = uuid },
            cancellationToken);
        return StatusCode(StatusCodes.Status204NoContent);
    }
}
